from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3_test.tests.boa_test import BoaTest


class TestLogical(BoaTest):

    def test_boolean_and(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.BOOLAND
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/BoolAnd.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_boolean_or(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.BOOLOR
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/BoolOr.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_boolean_not(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.NOT
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/BoolNot.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_sequence_boolean_or(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x03'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.BOOLOR
            + Opcode.LDARG2
            + Opcode.BOOLOR
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/BoolOrThreeElements.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mismatched_type_binary_operation(self):
        path = '%s/boa3_test/test_sc/logical_test/MismatchedOperandAnd.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_mismatched_type_unary_operation(self):
        path = '%s/boa3_test/test_sc/logical_test/MismatchedOperandNot.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_mixed_operations(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x03'
            + Opcode.LDARG0
            + Opcode.NOT
            + Opcode.LDARG1
            + Opcode.LDARG2
            + Opcode.BOOLOR
            + Opcode.BOOLAND
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/MixedOperations.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_logic_and_with_bool_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.AND
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/LogicAndBool.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_logic_and_with_int_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.AND
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/LogicAndInt.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mismatched_type_logic_and(self):
        path = '%s/boa3_test/test_sc/logical_test/MismatchedOperandLogicAnd.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_logic_or_with_bool_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.OR
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/LogicOrBool.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_logic_or_with_int_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.OR
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/LogicOrInt.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mismatched_type_logic_or(self):
        path = '%s/boa3_test/test_sc/logical_test/MismatchedOperandLogicOr.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_logic_not_with_bool_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.INVERT
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/LogicNotBool.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_logic_not_with_int_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.INVERT
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/LogicNotInt.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mismatched_type_logic_not(self):
        path = '%s/boa3_test/test_sc/logical_test/MismatchedOperandLogicNot.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_logic_xor_with_bool_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.XOR
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/LogicXorBool.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_logic_xor_with_int_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.XOR
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/LogicXorInt.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mismatched_type_logic_xor(self):
        path = '%s/boa3_test/test_sc/logical_test/MismatchedOperandLogicXor.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_logic_left_shift(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.SHL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/LogicLeftShift.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mismatched_type_logic_left_shift(self):
        path = '%s/boa3_test/test_sc/logical_test/MismatchedOperandLogicLeftShift.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_logic_right_shift(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.SHR
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/LogicRightShift.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mismatched_type_logic_right_shift(self):
        path = '%s/boa3_test/test_sc/logical_test/MismatchedOperandLogicRightShift.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)



from .base import IDBImplementation
from .cache import (Trackable,
                    TrackState,
                    CachedTXAccess,
                    CloneTXCache,
                    CachedBlockAccess,
                    CloneBlockCache,
                    CachedContractAccess,
                    CloneContractCache,
                    CachedStorageAccess,
                    CloneStorageCache,
                    AttributeCache)
from .contractstate import ContractState
from .snapshot import CloneSnapshot, Snapshot
from .storageitem import StorageItem
from .storagekey import StorageKey



import hashlib
from enum import IntEnum
from typing import List

from boa3.neo3 import settings
from boa3.neo3.core import Size as s, serialization, utils, types
from boa3.neo3.network import payloads
from boa3.neo3.vm import VMState


class TransactionAttributeUsage(IntEnum):
    URL = 0x81


class TransactionAttribute(serialization.ISerializable):
    """
    Attributes that can be attached to a Transaction.
    """

    def __init__(self, usage: TransactionAttributeUsage = None, data: bytes = None):
        self.usage = usage
        self.data = data if data else bytearray()

    def __len__(self):
        return s.uint8 + utils.get_var_size(self.data)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint8(self.usage)
        writer.write_var_bytes(self.data)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.usage = TransactionAttributeUsage(reader.read_uint8())
        self.data = reader.read_var_bytes(max=252)


class Transaction(serialization.ISerializable, payloads.IInventory):
    """
    Data to be executed by the NEO virtual machine.
    """
    #: the maximum number of bytes a single transaction may consists of
    MAX_TRANSACTION_SIZE = 102400
    #: the maximum time a transaction will be valid from height of creation plus this value.
    MAX_VALID_UNTIL_BLOCK_INCREMENT = 2102400
    #: the maximum number of transaction attributes for a single transaction
    MAX_TRANSACTION_ATTRIBUTES = 16
    #: the maximum number of cosigners for a single
    MAX_COSIGNERS = 16

    def __init__(self,
                 version: int = 0,
                 nonce: int = 0,
                 sender: types.UInt160 = None,
                 system_fee: int = 0,
                 network_fee: int = 0,
                 valid_until_block: int = 0,
                 attributes: List[TransactionAttribute] = None,
                 cosigners: List[payloads.Cosigner] = None,
                 script: bytes = None,
                 witnesses: List[payloads.Witness] = None,
                 protocol_magic: int = None):
        self.version = version
        self.nonce = nonce
        #: Script hash of the first signing authority
        self.sender = sender if sender else types.UInt160.zero()
        self.system_fee = system_fee
        self.network_fee = network_fee
        self.valid_until_block = valid_until_block
        self.attributes = attributes if attributes else []
        #: A list of authorities used by the :func:`ChecKWitness` smart contract system call.
        self.cosigners = cosigners if cosigners else []
        self.script = script if script else b''
        #: A list of signing authorities used to validate the transaction.
        self.witnesses = witnesses if witnesses else []

        # unofficial attributes
        self.vm_state = VMState.NONE
        self.block_height = 0
        #: The network protocol magic number to use in the Transaction hash function. Defaults to 0x4F454E
        #: Warning: changing this will change the TX hash which can result in dangling transactions in the database as
        #: deletion and duplication checking will fail.
        if protocol_magic:
            self.protocol_magic = protocol_magic
        elif settings.network.magic is not None:
            self.protocol_magic = settings.network.magic
        else:
            self.protocol_magic = 0x4F454E

    def __len__(self):
        return (s.uint8 + s.uint32 + s.uint160 + s.uint64 + s.uint64 + s.uint32
                + utils.get_var_size(self.attributes)
                + utils.get_var_size(self.cosigners)
                + utils.get_var_size(self.script)
                + utils.get_var_size(self.witnesses))

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.hash() != other.hash():
            return False
        return True

    def __hash__(self):
        # the TX hash() is a UInt256, we need to return an int
        # so we call hash() on the UInt256
        return hash(self.hash())

    def __deepcopy__(self, memodict={}):
        # not the best, but faster than letting deepcopy() do introspection
        with serialization.BinaryWriter() as bw:
            self.serialize_special(bw)
            with serialization.BinaryReader(bw.to_array()) as br:
                tx = Transaction()
                tx.deserialize_special(br)
                return tx

    def hash(self) -> types.UInt256:
        """
        Get a unique block identifier based on the unsigned data portion of the object.
        """
        with serialization.BinaryWriter() as bw:
            bw.write_uint32(self.protocol_magic)
            self.serialize_unsigned(bw)
            data_to_hash = bytearray(bw._stream.getvalue())
            data = hashlib.sha256(hashlib.sha256(data_to_hash).digest()).digest()
            return types.UInt256(data=data)

    @property
    def inventory_type(self) -> payloads.InventoryType:
        """
        Inventory type identifier.
        """
        return payloads.InventoryType.TX

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        self.serialize_unsigned(writer)
        writer.write_serializable_list(self.witnesses)

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        writer.write_uint8(self.version)
        writer.write_uint32(self.nonce)
        writer.write_serializable(self.sender)
        writer.write_int64(self.system_fee)
        writer.write_int64(self.network_fee)
        writer.write_uint32(self.valid_until_block)
        writer.write_serializable_list(self.attributes)
        writer.write_serializable_list(self.cosigners)
        writer.write_var_bytes(self.script)

    def serialize_special(self, writer: serialization.BinaryWriter) -> None:
        self.serialize(writer)
        writer.write_uint8(self.vm_state)
        writer.write_uint32(self.block_height)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.deserialize_unsigned(reader)
        self.witnesses = reader.read_serializable_list(payloads.Witness)

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        self.version = reader.read_uint8()
        if self.version > 0:
            raise ValueError("Deserialization error - invalid version")
        self.nonce = reader.read_uint32()
        self.sender = reader.read_serializable(types.UInt160)
        self.system_fee = reader.read_int64()
        if self.system_fee < 0:
            raise ValueError("Deserialization error - negative system fee")
        self.network_fee = reader.read_int64()
        if self.network_fee < 0:
            raise ValueError("Deserialization error - negative network fee")
        # Impossible overflow, only applicable to the C# implementation where they use longs
        # if (self.system_fee + self.network_fee < self.system_fee):
        #     raise ValueError("Deserialization error - overflow")
        self.valid_until_block = reader.read_uint32()
        self.attributes = reader.read_serializable_list(TransactionAttribute)
        self.cosigners = reader.read_serializable_list(payloads.Cosigner)
        self.script = reader.read_var_bytes(max=65535)
        if len(self.script) == 0:
            raise ValueError("Deserialization error - invalid script length 0")

    def deserialize_special(self, reader: serialization.BinaryReader) -> None:
        self.deserialize(reader)
        self.vm_state = VMState(reader.read_uint8())
        self.block_height = reader.read_uint32()

    def fee_per_byte(self) -> int:
        """
        Calculates the network fee per byte.

        Fee per byte = the TX's networkfee / TX's size

        Warning:
            Should only be called once the transaction is completely build and will no longer be modified.
        """
        return self.network_fee // len(self)

    def from_replica(self, replica):
        self.version = replica.version
        self.nonce = replica.nonce
        self.sender = replica.sender
        self.system_fee = replica.system_fee
        self.network_fee = replica.network_fee
        self.valid_until_block = replica.valid_until_block
        self.attributes = replica.attributes
        self.cosigners = replica.cosigners
        self.script = replica.script
        self.witnesses = replica.witnesses
        self.block_height = replica.block_height
        self.vm_state = replica.vm_state

    # TODO: implement Verify methods once we have Snapshot support



from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes, NotSupportedOperation, UnexpectedArgument, UnfilledArgument
from boa3.model.type.type import Type
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestVariable(BoaTest):

    # region TestLen

    def test_len_of_tuple(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.PUSH3      # a = (1, 2, 3)
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # b = len(a)
            + Opcode.SIZE
            + Opcode.STLOC1
            + Opcode.LDLOC1     # return b
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/LenTuple.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_len_of_list(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3      # array length
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # b = len(a)
            + Opcode.SIZE
            + Opcode.STLOC1
            + Opcode.LDLOC1     # return b
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/LenList.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_len_of_str(self):
        input = 'just a test'
        byte_input = String(input).to_bytes()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1            # push the bytes
            + Integer(len(byte_input)).to_byte_array()
            + byte_input
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.SIZE
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/LenString.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_len_of_bytes(self):
        byte_input = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.PUSHDATA1            # push the bytes
            + Integer(len(byte_input)).to_byte_array()
            + byte_input
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.SIZE
            + Opcode.STLOC1
            + Opcode.LDLOC1
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/LenBytes.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_len_of_no_collection(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/LenMismatchedType.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_len_too_many_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/LenTooManyParameters.py' % self.dirname
        self.assertCompilerLogs(UnexpectedArgument, path)

    def test_len_too_few_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/LenTooFewParameters.py' % self.dirname
        self.assertCompilerLogs(UnfilledArgument, path)

    # endregion

    # region TestAppend

    def test_append_tuple(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/AppendTuple.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_append_sequence(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/AppendSequence.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_append_mutable_sequence(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x02'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # a.append(4)
            + Opcode.PUSH4
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.JMP
                + Integer(3).to_byte_array(min_length=1)
                + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/AppendMutableSequence.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_append_mutable_sequence_with_builtin(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x02'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # a.append(4)
            + Opcode.PUSH4
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.JMP
                + Integer(3).to_byte_array(min_length=1)
                + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/AppendMutableSequenceBuiltinCall.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_append_too_many_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/AppendTooManyParameters.py' % self.dirname
        self.assertCompilerLogs(UnexpectedArgument, path)

    def test_append_too_few_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/AppendTooFewParameters.py' % self.dirname
        self.assertCompilerLogs(UnfilledArgument, path)

    # endregion

    # region TestClear

    def test_clear_tuple(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ClearTuple.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_clear_mutable_sequence(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x02'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # a.clear()
                + Opcode.DUP
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(9).to_byte_array(min_length=1)
                + Opcode.DROP
                + Opcode.PUSHDATA1
                + Integer(0).to_byte_array(min_length=1)
                + Opcode.CONVERT
                + Type.bytearray.stack_item
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.CLEARITEMS
                + Opcode.JMP
                + Integer(3).to_byte_array(min_length=1)
                + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/ClearMutableSequence.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_clear_mutable_sequence_with_builtin(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x02'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # MutableSequence.clear(a)
                + Opcode.DUP
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(9).to_byte_array(min_length=1)
                + Opcode.DROP
                + Opcode.PUSHDATA1
                + Integer(0).to_byte_array(min_length=1)
                + Opcode.CONVERT
                + Type.bytearray.stack_item
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.CLEARITEMS
                + Opcode.JMP
                + Integer(3).to_byte_array(min_length=1)
                + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/ClearMutableSequenceBuiltinCall.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_clear_too_many_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ClearTooManyParameters.py' % self.dirname
        self.assertCompilerLogs(UnexpectedArgument, path)

    def test_clear_too_few_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ClearTooFewParameters.py' % self.dirname
        self.assertCompilerLogs(UnfilledArgument, path)

    # endregion

    # region TestReverse

    def test_reverse_tuple(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ReverseTuple.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_reverse_mutable_sequence(self):
        expected_output = (
                Opcode.INITSLOT     # function signature
                + b'\x01'
                + b'\x00'
                + Opcode.PUSH3      # a = [1, 2, 3]
                + Opcode.PUSH2
                + Opcode.PUSH1
                + Opcode.PUSH3
                + Opcode.PACK
                + Opcode.STLOC0
                + Opcode.LDLOC0     # a.reverse()
                + Opcode.REVERSEITEMS
                + Opcode.LDLOC0     # return a
                + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/ReverseMutableSequence.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_reverse_mutable_sequence_with_builtin(self):
        data = b'\x01\x02\x03'
        expected_output = (
                Opcode.INITSLOT     # function signature
                + b'\x01'
                + b'\x00'
                + Opcode.PUSHDATA1  # a = bytearray(b'\x01\x02\x03')
                + Integer(len(data)).to_byte_array(min_length=1)
                + data
                + Opcode.CONVERT
                + Type.bytes.stack_item
                + Opcode.STLOC0
                + Opcode.LDLOC0     # MutableSequence.reverse(a)
                + Opcode.REVERSEITEMS
                + Opcode.LDLOC0     # return a
                + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/ReverseMutableSequenceBuiltinCall.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_reverse_too_many_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ReverseTooManyParameters.py' % self.dirname
        self.assertCompilerLogs(UnexpectedArgument, path)

    def test_reverse_too_few_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ReverseTooFewParameters.py' % self.dirname
        self.assertCompilerLogs(UnfilledArgument, path)

    # endregion

    # region TestExtend

    def test_extend_tuple(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ExtendTuple.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_extend_sequence(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ExtendSequence.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_extend_mutable_sequence(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x02'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # a.extend((4, 5, 6))
            + Opcode.PUSH6      # (4, 5, 6)
            + Opcode.PUSH5
            + Opcode.PUSH4
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.UNPACK     # a.extend
                + Opcode.JMP
                + Integer(9).to_byte_array(signed=True, min_length=1)
                + Opcode.DUP
                + Opcode.INC
                + Opcode.PICK
                + Opcode.PUSH2
                + Opcode.ROLL
                + Opcode.APPEND
                + Opcode.DEC
                + Opcode.DUP
                + Opcode.JMPIF
                + Integer(-8).to_byte_array(signed=True, min_length=1)
                + Opcode.DROP
                + Opcode.DROP
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/ExtendMutableSequence.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_extend_mutable_sequence_with_builtin(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x02'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # MutableSequence.extend(a, [4, 5, 6])
            + Opcode.PUSH6      # [4, 5, 6]
            + Opcode.PUSH5
            + Opcode.PUSH4
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.UNPACK     # a.extend
                + Opcode.JMP
                + Integer(9).to_byte_array(signed=True, min_length=1)
                + Opcode.DUP
                + Opcode.INC
                + Opcode.PICK
                + Opcode.PUSH2
                + Opcode.ROLL
                + Opcode.APPEND
                + Opcode.DEC
                + Opcode.DUP
                + Opcode.JMPIF
                + Integer(-8).to_byte_array(signed=True, min_length=1)
                + Opcode.DROP
                + Opcode.DROP
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/built_in_methods_test/ExtendMutableSequenceBuiltinCall.py' % self.dirname

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_extend_too_many_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ExtendTooManyParameters.py' % self.dirname
        self.assertCompilerLogs(UnexpectedArgument, path)

    def test_extend_too_few_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ExtendTooFewParameters.py' % self.dirname
        self.assertCompilerLogs(UnfilledArgument, path)

    # endregion

    # region TestToScriptHash

    def test_script_hash_int(self):
        from boa3.neo import to_script_hash
        script_hash = to_script_hash(Integer(123).to_byte_array())
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(script_hash)).to_byte_array(min_length=1)
            + script_hash
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/built_in_methods_test/ScriptHashInt.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_script_hash_int_with_builtin(self):
        from boa3.neo import to_script_hash
        script_hash = to_script_hash(Integer(123).to_byte_array())
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(script_hash)).to_byte_array(min_length=1)
            + script_hash
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/built_in_methods_test/ScriptHashIntBuiltinCall.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_script_hash_str(self):
        from boa3.neo import to_script_hash
        script_hash = to_script_hash(String('123').to_bytes())
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(script_hash)).to_byte_array(min_length=1)
            + script_hash
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/built_in_methods_test/ScriptHashStr.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_script_hash_str_with_builtin(self):
        from boa3.neo import to_script_hash
        script_hash = to_script_hash(String('123').to_bytes())
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(script_hash)).to_byte_array(min_length=1)
            + script_hash
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/built_in_methods_test/ScriptHashStrBuiltinCall.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_script_hash_variable(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ScriptHashVariable.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_script_hash_variable_with_builtin(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ScriptHashVariableBuiltinCall.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_script_hahs_too_many_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ScriptHashTooManyParameters.py' % self.dirname
        self.assertCompilerLogs(UnexpectedArgument, path)

    def test_script_hash_too_few_parameters(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ScriptHashTooFewParameters.py' % self.dirname
        self.assertCompilerLogs(UnfilledArgument, path)

    def test_script_hash_mismatched_types(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ScriptHashMismatchedType.py' % self.dirname
        # TODO: change to MismatchedTypes when 'to_script_hash' with variables is implemented
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_script_hash_builtin_mismatched_types(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ScriptHashBuiltinMismatchedType.py' % self.dirname
        # TODO: change to MismatchedTypes when 'to_script_hash' with variables is implemented
        self.assertCompilerLogs(NotSupportedOperation, path)

    # endregion

    # region TestToBytes

    def test_int_to_bytes(self):
        value = Integer(123).to_byte_array()
        expected_output = (
            Opcode.PUSHDATA1        # (123).to_bytes()
            + Integer(len(value)).to_byte_array(min_length=1)
            + value
            + Opcode.CONVERT
            + Type.int.stack_item
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/built_in_methods_test/IntToBytes.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_int_to_bytes_with_builtin(self):
        value = Integer(123).to_byte_array()
        expected_output = (
            Opcode.PUSHDATA1        # int.to_bytes(123)
            + Integer(len(value)).to_byte_array(min_length=1)
            + value
            + Opcode.CONVERT
            + Type.int.stack_item
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/built_in_methods_test/IntToBytesWithBuiltin.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_str_to_bytes(self):
        value = String('123').to_bytes()
        expected_output = (
            Opcode.PUSHDATA1        # '123'.to_bytes()
            + Integer(len(value)).to_byte_array(min_length=1)
            + value
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/built_in_methods_test/StrToBytes.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_str_to_bytes_with_builtin(self):
        value = String('123').to_bytes()
        expected_output = (
            Opcode.PUSHDATA1        # str.to_bytes('123')
            + Integer(len(value)).to_byte_array(min_length=1)
            + value
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/built_in_methods_test/StrToBytesWithBuiltin.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_to_bytes_mismatched_types(self):
        path = '%s/boa3_test/test_sc/built_in_methods_test/ToBytesMismatchedType.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    # endregion



import ast

from boa3.neo.vm.VMCode import VMCode


class DebugInstruction:
    def __init__(self, code: VMCode, start_line: int, start_col: int, end_line: int = None, end_col: int = None):
        if end_line is None or end_line < start_line:
            end_line = start_line
        if end_col is None or end_col < end_col:
            end_col = start_col

        self.code: VMCode = code
        self.start_line: int = start_line
        self.start_col: int = start_col
        self.end_line: int = end_line
        self.end_col: int = end_col

    def __str__(self) -> str:
        return '{0}:{1}-{2}:{3} {4}'.format(self.start_line, self.start_col,
                                            self.end_line, self.end_col,
                                            self.code)

    @classmethod
    def build(cls, ast_node: ast.AST, bytecode: VMCode):
        end_line: int = ast_node.end_lineno if hasattr(ast_node, 'end_lineno') else ast_node.lineno
        end_col: int = ast_node.end_col_offset if hasattr(ast_node, 'end_lineno') else ast_node.col_offset
        return cls(bytecode, ast_node.lineno, ast_node.col_offset, end_line, end_col)



from typing import Any, Dict, List, Tuple


def public(*args):
    """
    This decorator identifies which methods should be included in the abi file
    """
    def public_wrapper():
        pass
    return public_wrapper


def metadata(*args):
    """
    This decorator identifies the function that returns the metadata object of the smart contract
    This can be used to only one function. Using this decorator in multiple functions will raise a compiler error.
    """
    def metadata_wrapper():
        pass
    return metadata_wrapper


def to_script_hash(data_bytes: Any) -> bytes:
    """
    Converts a data to a script hash.

    :param data_bytes: data to hash.
    :return: the script hash of the data
    :rtype: bytes
    """
    pass


def Event(*args, **kwargs):
    pass


def CreateNewEvent(arguments: List[Tuple[str, type]] = [], event_name: str = '') -> Event:
    """
    Creates a new event.

    :param arguments: the list of the events args' names and types
    :param event_name: custom name of the event. It's filled with the variable name if not specified.
    :return: the new event
    :rtype: Event
    """
    pass


class NeoMetadata:
    """
    This class stores the smart contract manifest information.

    :ivar has_storage: is a boolean value indicating whether the contract has a storage. False by default.
    :type has_storage: bool
    :ivar is_payable: is a boolean value indicating whether the contract accepts transfers. False by default.
    :type is_payable: bool
    :ivar author: the smart contract author. None by default;
    :type author: str or None
    :ivar email: the smart contract author email. None by default;
    :type email: str or None
    :ivar description: the smart contract description. None by default;
    :type description: str or None
    """

    def __init__(self):
        from typing import Optional

        # features
        self.has_storage: bool = False
        self.is_payable: bool = False

        # extras
        self.author: Optional[str] = None
        self.email: Optional[str] = None
        self.description: Optional[str] = None

    @property
    def extra(self) -> Dict[str, Any]:
        """
        Gets the metadata extra information

        :return: a dictionary that maps each extra value with its name. Empty by default.
        """
        extra = {}
        if isinstance(self.author, str):
            extra['Author'] = self.author
        if isinstance(self.email, str):
            extra['Email'] = self.email
        if isinstance(self.description, str):
            extra['Description'] = self.description
        return extra



import ast
from typing import Optional

from boa3.model.expression import IExpression
from boa3.model.method import Method
from boa3.model.type.itype import IType


class Property(IExpression):
    """
    A class used to represent a variable

    :ivar var_type: the type of the variable.
    """

    def __init__(self, getter: Method, setter: Method = None, origin_node: Optional[ast.AST] = None):
        super().__init__(origin_node)
        self._getter: Method = getter
        self._setter: Optional[Method] = setter

    @property
    def shadowing_name(self) -> str:
        return 'property'

    @property
    def type(self) -> IType:
        return self._getter.type

    @property
    def getter(self) -> Method:
        return self._getter

    @property
    def setter(self) -> Optional[Method]:
        return self._setter

    def __str__(self) -> str:
        return str(self.type)



from abc import ABC
from typing import Any, Iterable, Set

from boa3.model.type.collection.icollection import ICollectionType
from boa3.model.type.itype import IType
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.StackItemType import StackItemType


class MappingType(ICollectionType, ABC):
    """
    An interface used to represent Python mapping type
    """

    def __init__(self, identifier: str, keys_type: Set[IType], values_type: Set[IType]):
        super().__init__(identifier, keys_type=keys_type, values_type=values_type)

    @property
    def identifier(self) -> str:
        return '{0}[{1}, {2}]'.format(self._identifier, self.key_type.identifier, self.value_type.identifier)

    @property
    def value_type(self) -> IType:
        return self.item_type

    @property
    def default_value(self) -> Any:
        return {}

    def is_type_of(self, value: Any) -> bool:
        if self._is_type_of(value):
            if isinstance(value, MappingType):
                return (self.key_type.is_type_of(value.key_type)
                        and self.value_type.is_type_of(value.value_type))
            return True
        return False

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Map

    @property
    def stack_item(self) -> StackItemType:
        return StackItemType.Map

    def is_valid_key(self, key_type: IType) -> bool:
        return self.valid_key.is_type_of(key_type)

    @property
    def valid_key(self) -> IType:
        return self.key_type

    @classmethod
    def filter_types(cls, values_type) -> Set[IType]:
        if values_type is None:
            values_type = set()
        elif not isinstance(values_type, set):
            if isinstance(values_type, Iterable):
                values_type = set(values_type)
            else:
                values_type = {values_type}

        if len(values_type) > 1 and all(isinstance(x, MappingType) for x in values_type):
            first_item: MappingType = list(values_type)[0]
            mapping_type = type(first_item)          # first mapping type

            k_types = set(value.key_type for value in values_type)
            v_types = set(value.value_type for value in values_type)

            if all(isinstance(x, mapping_type) for x in values_type):
                values_type = {mapping_type(keys_type=k_types, values_type=v_types)}
            else:
                from boa3.model.type.type import Type
                generic_type: IType = Type.get_generic_type(*values_type)
                if isinstance(generic_type, MappingType):
                    values_type = {generic_type.build_collection(k_types, v_types)}

            return values_type

        # if any value is not a map, call the collection filter
        return super().filter_types(values_type)

    @classmethod
    def build(cls, value: Any):
        if cls._is_type_of(value):
            if isinstance(value, dict):
                keys = list(value.keys())
                values = list(value.values())
            else:
                keys = value.key_type
                values = value.value_type

            keys_types: Set[IType] = cls.get_types(keys)
            values_types: Set[IType] = cls.get_types(values)
            return cls(keys_types, values_types)



from abc import ABC, abstractmethod
from typing import Any, Iterable, Set, Union

from boa3.model.type.itype import IType


class ICollectionType(IType, ABC):
    """
    An interface used to represent Python mapping type
    """

    def __init__(self, identifier: str, keys_type: Set[IType] = None, values_type: Set[IType] = None):
        if keys_type is None:
            keys_type = set()
        self.key_type: IType = self._get_collection_type(keys_type)

        if values_type is None:
            values_type = set()
        self.item_type: IType = self._get_collection_type(values_type)
        super().__init__(identifier)

    @property
    def identifier(self) -> str:
        return '{0}[{1}]'.format(self._identifier, self.item_type.identifier)

    def is_type_of(self, value: Any) -> bool:
        if self._is_type_of(value):
            if isinstance(value, ICollectionType):
                return self.item_type.is_type_of(value.item_type)
            return True
        return False

    @abstractmethod
    def is_valid_key(self, key_type: IType) -> bool:
        """
        Verifies if the given type is valid to retrieve values

        :param key_type: type to validate.
        :return: A boolean value that represents if the value is valid.
        """
        return False

    @property
    def can_reassign_values(self) -> bool:
        return False

    @property
    @abstractmethod
    def valid_key(self) -> IType:
        pass

    def _get_collection_type(self, values_type: Set[IType]):
        if len(values_type) != 1:
            from boa3.model.type.anytype import anyType
            val_type: IType = anyType
        else:
            val_type: IType = list(values_type)[0]

        return val_type

    @classmethod
    def get_types(cls, value: Any) -> Set[IType]:
        from boa3.model.type.type import Type
        if isinstance(value, IType):
            return {value}

        if not isinstance(value, Iterable):
            value = {value}

        types: Set[IType] = {val if isinstance(val, IType) else Type.get_type(val) for val in value}
        return cls.filter_types(types)

    @classmethod
    def filter_types(cls, values_type) -> Set[IType]:
        if values_type is None:
            values_type = set()
        elif not isinstance(values_type, set):
            if isinstance(values_type, Iterable):
                values_type = set(values_type)
            else:
                values_type = {values_type}

        if len(values_type) > 1:
            from boa3.model.type.type import Type
            if any(t is Type.any or t is Type.none for t in values_type):
                return {Type.any}

            if any(not isinstance(x, ICollectionType) for x in values_type):
                return {Type.any}
            # verifies if all the types are the same collection with different arguments

            if not all(isinstance(x, ICollectionType) for x in values_type):
                return {Type.get_generic_type(*values_type)}
            else:
                first_item: ICollectionType = list(values_type)[0]
                collection_type = type(first_item)  # first sequence type
                value_type = type(first_item.item_type)  # first sequence values type

            if all(isinstance(x, collection_type) for x in values_type):
                # if all the types are the same sequence type, build this sequence with any as parameters
                types = set(value.item_type for value in values_type)
                values_type = {collection_type(values_type=types)}
            else:
                generic_type: IType = Type.get_generic_type(*values_type)
                if (isinstance(generic_type, ICollectionType)
                        and all(isinstance(x.item_type, value_type) for x in values_type)):
                    # the collections doesn't have the same type but the value type is the same
                    # for example: Tuple[int] and List[int]
                    values_type = {generic_type.build_collection(first_item.item_type)}
                else:
                    # otherwise, built a generic sequence with any as parameters
                    values_type = {generic_type}

        return values_type

    @classmethod
    def build_collection(cls, *value_type: Union[IType, Iterable]):
        """
        Creates a collection type instance with the given value

        :param value_type: value to build the type
        :type value_type: IType or Iterable
        :return: The built collection type.
        :rtype: IType or None
        """
        from inspect import signature
        # if the number of params is different than the constructor args (except self arg)
        # returns an object with the default values
        if len(value_type) != len(signature(cls.__init__).parameters) - 1:
            return cls()

        params = []
        for arg in value_type:
            if isinstance(arg, Iterable):
                argument = set(arg)
            else:
                argument = {arg}
            params.append(argument)

        return cls(*params)
